#include <errno.h>
#include <stdlib.h>
#include "pvmInt.h"
#include "patchlevel.h"

/*
 * Not all systems declare the errno variable in errno.h. so this
 * file does it explicitly.  The list of system error messages also
 * isn't generally declared in a header file anywhere.
 */

extern int errno;
/*
 * For each Pvm-event registered in a call to Pvm_CreatePvmHandler,
 * there is one record of the following type.  All of these records
 * are chained together into a single list.
 */

typedef struct PvmHandler {
    int tid;			/* task identifier (or -1 if none) */
    int msgtag;			/* message identifier (or -1 if none) */
    Pvm_PvmProc *proc;		/* procedure to call if pvm packet received */
    ClientData clientData;	/* Argument to pass to proc. */
    struct PvmHandler *nextPtr; /* Next in list */
} PvmHandler;
static PvmHandler *firstPvmHandlerPtr = NULL;  /* list of all Pvm handlers */

/*
 * The data structure below is used by the "pvm bind" command to remember
 * the command to be executed later.
 */

typedef struct PvmBindInfo {
    Tcl_Interp *interp;		/* Interpreter in which to execute command. */
    char *command;		/* Command to execute.  Malloc'ed, so must
				 * be freed when structure is deallocated. */
    int tid;			/* task identifier (or -1 if none) */
    int msgtag;			/* message identifier (or -1 if none) */
    struct PvmBindInfo *nextPtr;/* Next in list of all "pvm bind" commands for
				 * the application. */
} PvmBindInfo;

static PvmBindInfo *firstPvmBindPtr = NULL;
				/* First in list of all pending "pvm bind"
				 * commands. */


typedef struct PvmTidInfo {
    int tid;			/* task identifier */
    Tcl_Interp *interp;		/* interpreter that spawned the tid
				   (NULL if unknown) */
    FILE *output;		/* filehandle for output redirect */
    struct PvmTidInfo *nextPtr; /* next in list */
} PvmTidInfo;

static PvmTidInfo *firstPvmTidPtr = NULL;

static Tcl_Interp *Pvm_interp = NULL; /* interpreter that initialized
					 mytid for the first time*/
static int mytid = 0;
static int pvm_idle_flag = 0;	/* flag that indicates if the pvm idle-routine is triggered */
static int pvm_nfds = 0	;	/* number of file descriptors in use for pvm */
static int *pvm_fdlist = NULL;   /* list of file descriptors in use for pvm (malloced)*/

/*
 * Prototypes for procedures in this file:
 */

int Pvm_PvmCmd _ANSI_ARGS_((ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]));
int Pvm_ExitCmd _ANSI_ARGS_((ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]));
int Pvm_File_handle _ANSI_ARGS_((ClientData clientData, int mask));
Pvm_PvmProc Pvm_EvalCommand;


/*
 *--------------------------------------------------------------
 *
 * Pvm_StartDaemon --
 *	Start pvm daemon.
 *
 * Results:
 *	0 if failed.
 *	tid of current process if succeeded.
 *
 * Side effects:
 *	None
 *
 *--------------------------------------------------------------
 */
static char *pvmd_args = "~/.tkpvm_hosts";

int
Pvm_StartDaemon(interp)
    Tcl_Interp *interp;
{
    int m;
    if(mytid) return mytid;
/*  start PVM daemon, if not yet started */
    pvm_setopt(PvmAutoErr, 0);
    m = pvm_start_pvmd(1, &pvmd_args, 1);
    if (m < 0) {
	if (m != PvmDupHost) {
	    Tcl_AppendResult(interp,"Sorry, cannot start PVM daemon\n",(char *)NULL);
	    pvm_exit();
	    return 0;
	}
    }
    return mytid = pvm_mytid();
}

/*
 *--------------------------------------------------------------
 *
 * Pvm_Cleanup --
 *	This function is called as an Idle handler. It cleans
 *	up file handles that are not used any more
 *	probably because the associated processes are killed.
 *
 * Results:
 *	None
 *
 * Side effects:
 *	None
 *
 *--------------------------------------------------------------
 */

static void
Pvm_Cleanup(clientData)
    ClientData clientData;	/* not used */
{
    int *fdlist, i;

    /*
     * remove all file handles in use by pvm. If there is
     * only one, this must be the connection with the local
     * daemon, which should be never deleted.
     */

    if(pvm_nfds>1) {
	for(i=0; i<pvm_nfds; i++) {
	    Tk_DeleteFileHandler(pvm_fdlist[i]);
	}
	ckfree(pvm_fdlist);
    }
    pvm_nfds = 0;

    /*
     * register File handler for all fd's in use for pvm, such
     * that it will be executed when a PVM-package is received.
     */

    pvm_nfds = pvm_getfds(&fdlist);
    for(i=0; i<pvm_nfds;i++) {
	Tk_CreateFileHandler(fdlist[i],TK_READABLE,
	    (Tk_FileProc *)Pvm_File_handle,(ClientData)NULL);
    }
    /*
     * If the number of file handles is bigger than 1 copy
     * the list to another array, so the handles can be
     * deleted later when no longer in use. If there is
     * only one, this is not necessary (see above).
     */

    if(pvm_nfds>1) {
	pvm_fdlist = ckalloc(pvm_nfds * sizeof(int));
	memcpy(pvm_fdlist, fdlist, pvm_nfds*sizeof(int));
    }
    pvm_idle_flag = 0;
}

/*
 *--------------------------------------------------------------
 *
 * Pvm_InitPvmCmd --
 *	Create 'pvm' command.
 *
 * Results:
 *	None
 *
 * Side effects:
 *	Starts PVM daemon, if possible and if not already
 *	started.
 *
 *--------------------------------------------------------------
 */
int
Pvm_InitPvmCmd(interp)
    Tcl_Interp *interp;
{
    char * libDir;
    char buf[10];
    Pvm_interp = interp;

    Tcl_CreateCommand(interp, "pvm", Pvm_PvmCmd, (ClientData) NULL,
	    (void (*)()) NULL);
    if(Tcl_Eval(interp,"set tk_version")!=TCL_OK) {
	Tcl_ResetResult(interp);
	Pvm_EventInit(interp);
    }
    /*
     * Set variables for the intepreter.
     */

    libDir = getenv("TKPVM_LIBRARY");
    if (libDir == NULL) {
	libDir = TKPVM_LIBRARY;
    }
    Tcl_SetVar(interp, "pvm_version",     pvm_version(),     TCL_GLOBAL_ONLY);
    Tcl_SetVar(interp, "tkpvm_library",   libDir,            TCL_GLOBAL_ONLY);
    Tcl_SetVar(interp, "tkpvm_patchLevel",TKPVM_PATCH_LEVEL, TCL_GLOBAL_ONLY);
    Tcl_SetVar(interp, "tkpvm_version",   TKPVM_VERSION,     TCL_GLOBAL_ONLY);

}

/*
 *--------------------------------------------------------------
 *
 * Pvm_Init --
 *	Create 'pvm' command and execute startup script
 *
 * Results:
 *	None
 *
 * Side effects:
 *	None
 *
 *--------------------------------------------------------------
 */

int
Pvm_Init(interp)
    Tcl_Interp *interp;
{
    static char initCmd[] =
	"if [file exists $tkpvm_library/pvm.tcl] {\n\
	    source $tkpvm_library/pvm.tcl\n\
	} else {\n\
	    set msg \"can't find $tkpvm_library/pvm.tcl; perhaps you \"\n\
	    append msg \"need to\\ninstall tkPvm or set your TKPVM_LIBRARY \"\n\
	    append msg \"environment variable?\"\n\
	    error $msg\n\
	}";

    Pvm_InitPvmCmd(interp);

    return Tcl_Eval(interp, initCmd);

}


/*
 *--------------------------------------------------------------
 *
 * Pvm_ReturnError --
 *
 *	This procedure changes pvm errors into Tcl errors.
 *
 *
 * Results:
 *	TCL_OK if there is no error.
 *
 *	TCL_ERROR if there is an error. Then interp will
 *	contain the PVM error message.
 *
 * Side effects:
 *	None
 *
 *
 *--------------------------------------------------------------
 */

int
Pvm_ReturnError(interp, error)
    Tcl_Interp *interp;		/* Current interpreter. */
    int error;			/* PVM error-number */
{
    char storage[4];
    static char *PVM_ERROR_MSG[] = {NULL,
				    NULL,
				    "bad parameter",
				    "barrier count mismatch",
				    NULL,
				    "read past end of buffer",
				    "no such host",
				    "no such executable",
				    NULL,
				    NULL,
				    "can't get memory",
				    NULL,
				    "can't decode received msg",
				    NULL,
				    "pvmd not responding",
				    "no current buffer",
				    "bad message id",
				    "null group name is illegal",
				    "already in group",
				    "no group with that name",
				    "not in group",
				    "no such instance in group",
				    "host failed",
				    "no parent task",
				    "function not implemented",
				    "pvmd system error",
				    "pvmd-pvmd protocol mismatch",
				    "out of resources",
				    "host already configured",
				    "failed to exec new slave pvmd",
				    "already doing operation",
				    "no such task",
				    "no such (group, instance)",
				    "(group,instance) already exists"};
    if(error>=0) return TCL_OK;
    error = -error;
    if(error>1) Tcl_AppendResult(interp,"PVM error: ",(char *)NULL);
    if(error<34 && PVM_ERROR_MSG[error]) {
	Tcl_AppendResult(interp, PVM_ERROR_MSG[error],(char *)NULL);
    } else {
	sprintf(storage,"-%d",error);
	Tcl_AppendResult(interp, "unknown error: ",storage,(char *)NULL);
    }
    return TCL_ERROR;
}


/*
 *--------------------------------------------------------------
 *
 * Pvm_PvmCmd --
 *
 *	This procedure is invoked to process the "pvm" Tcl
 *	command.
 *
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	If the pvm daemon is not yet started, this will be done.
 *
 *
 *--------------------------------------------------------------
 */

int
Pvm_PvmCmd(clientData, interp, argc, argv)
    ClientData clientData;	/* Main window associated with
				 * interpreter. */
    Tcl_Interp *interp;		/* Current interpreter. */
    int argc;			/* Number of arguments. */
    char **argv;		/* Argument strings. */
{
    char storage[1024];

    union value {	char c;
			int i;
			float f;
			double d;
    } v;
    int tid, msgtag, pvmerror=-1;

    if (!mytid) {
	if(argc>1 && !strcmp(argv[1],"exit")) {
	    return TCL_OK;
	}
	if(!Pvm_StartDaemon(interp))
	    return TCL_ERROR;
    }

    if (!pvm_idle_flag) {
	Tk_DoWhenIdle(Pvm_Cleanup,(ClientData) NULL);
	pvm_idle_flag = 1;
    }

    if(argc<2) {
	Tcl_AppendResult(interp, "Option missing for \"", argv[0],
		"\"-command: should be addhost, bind, delhost, exit, halt, joingroup, kill, leavegroup,",
		" mytid, parent, recv, reset, send, spawn or tasks",(char *) NULL);
    } else if(!strcmp(argv[1],"bind")) {
	char *command;
	PvmBindInfo *prev_pevPtr, *pevPtr;
	PvmTidInfo *infoPtr;
	if(argc<4) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
				" bind tid msgtag command\"",(char *) NULL);
	    goto pvmend;
	}
	if (!strcmp(argv[2],"parent")) {
	    tid = pvm_parent();
	    if(tid==PvmSysErr||tid==PvmNoParent) {
		pvmerror = tid;
		goto pvmend;
	    }
	} else if (!strcmp(argv[2],"id")) {
	    tid = pvm_mytid();
	    if(tid==PvmSysErr) {
		pvmerror = tid;
		goto pvmend;
	    }
	} else if(!strcmp(argv[2],"any")) {
	    tid = -1;
	} else if (Tcl_GetInt(interp, argv[2], &tid) != TCL_OK) {
	    goto pvmend;
	} else if(tid<0x40000) {
	    pvmerror = PvmBadParam;
	    goto pvmend;
	}
	if (argc>4 && *argv[4]) {
	    infoPtr = firstPvmTidPtr;
	    while(infoPtr && infoPtr->tid !=tid) {
		infoPtr = infoPtr->nextPtr;
	    }
	    if(!infoPtr) {
		if((pvmerror=pvm_notify(PvmTaskExit,KILLTID_MSGTAG, 1, &tid))<0) {
		    goto pvmend;
		}
		infoPtr = (PvmTidInfo *) ckalloc(sizeof(PvmTidInfo));
		infoPtr->tid = tid;
		infoPtr->interp = interp;
		infoPtr->nextPtr = firstPvmTidPtr;
		firstPvmTidPtr = infoPtr;
	    }
	}
	pvmerror = -1;
	if(!strcmp(argv[3],"any"))
	    msgtag = -1;
	else if(!strcmp(argv[3],"kill"))
	    msgtag = -2;
	else if (Tcl_GetInt(interp, argv[3], &msgtag) != TCL_OK) {
	    goto pvmend;
	} else if (msgtag<0&&msgtag>-3) {
	    Tcl_AppendResult(interp,"The value \"",argv[3],
				"\" is reserved for internal use",(char *) NULL);
	    goto pvmend;
	}
	if (argc < 5) {
	    command = (char *)NULL;
	} else if (argc == 5) {
	    command = (char *) ckalloc((unsigned)(strlen(argv[4])+1));
	    strcpy(command,argv[4]);
	} else {
	    command = Tcl_Concat(argc-4, argv+4);
	}

	/*
	 * Locate an existing file handler for this file, if one exists,
	 * and make a new one if none currently exists.
	 */
	prev_pevPtr = NULL;
	for (pevPtr = firstPvmBindPtr; ; prev_pevPtr = pevPtr, pevPtr = pevPtr->nextPtr) {
	    if (pevPtr == NULL) {
		if (argc == 4) {
		    pvmerror = 0;
		    goto pvmend;
		}
		pevPtr = (PvmBindInfo *) ckalloc(sizeof(PvmBindInfo));
		pevPtr->interp = interp;
		pevPtr->tid = tid;
		pevPtr->msgtag = msgtag;
		pevPtr->command = command;
		pevPtr->nextPtr = firstPvmBindPtr;
		prev_pevPtr = NULL;
		firstPvmBindPtr = pevPtr;
		break;
	    }
	    if (pevPtr->tid == tid && pevPtr->msgtag == msgtag) {
		break;
	    }
	}
	if(command==NULL) {
	    Tcl_AppendResult(interp, pevPtr->command, (char *)NULL);
	    pvmerror = 0;
	    goto pvmend;
	} else if(*command=='\0') {
	    Pvm_DeletePvmHandler(tid ,msgtag);
	    if(pevPtr->command) {
		ckfree(pevPtr->command);
	    }
	    if(prev_pevPtr == NULL) {
		firstPvmBindPtr = firstPvmBindPtr->nextPtr;
	    } else {
		prev_pevPtr->nextPtr = pevPtr->nextPtr;
	    }
	    ckfree(pevPtr);
	} else {
	    Pvm_CreatePvmHandler(tid ,msgtag ,Pvm_EvalCommand
						,(ClientData)pevPtr);
	}
        pvmerror = 0;
    } else if(!strcmp(argv[1],"spawn")) {
	int flag=0,ntask=1,*tids,*t;
	char *where = NULL,*process= NULL; FILE *output= NULL;
	if(argc<3) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
				" spawn ?options? process\"",(char *) NULL);
	    goto pvmend;
	}
	argc-=2; argv+=2;
	while(argc>1 && **argv=='-') {
	    if(!strcmp(*argv,"-ntask")) {
		if(Tcl_GetInt(interp, argv[1],&ntask) != TCL_OK) {
		    return TCL_ERROR;
		}
	    } else if(!strcmp(*argv,"-host")) {
		if(!*argv[1]||!strcmp(argv[1],"any")) {
		    flag &= ~PvmTaskHost;
		} else {
		    where = argv[1]; flag = (flag&(~PvmTaskArch))|PvmTaskHost;
		    if(*argv[1]=='!'||*argv[1]=='~') {
			flag &= ~PvmHostCompl; where++;
		    } else {
			flag |= PvmHostCompl;
		    }
		}
	    } else if(!strcmp(*argv,"-output")) {
		char *p;
		if (Tcl_GetOpenFile(interp,argv[1], 1, 1, &output) != TCL_OK) {
		    return TCL_ERROR;
		}
	    } else if(!strcmp(*argv,"-arch")&& !(flag&PvmTaskHost)) {
		if(!*argv[1]||!strcmp(argv[1],"any")) {
		    flag &= ~PvmTaskArch;
		} else {
		    where = argv[1]; flag |= PvmTaskArch;
		    if(*argv[1]=='!'||*argv[1]=='~') {
			flag &= ~PvmHostCompl; where++;
		    } else {
			flag |= PvmHostCompl;
		    }
		}
	    } else if(!strcmp(*argv,"-debug")) {
		if(Tcl_GetBoolean(interp, argv[1],&v.i) != TCL_OK) {
		    return TCL_ERROR;
		}
		if(v.i) flag |= PvmTaskDebug; else flag &= ~PvmTaskDebug;
	    } else if(!strcmp(*argv,"-trace")) {
		if(Tcl_GetBoolean(interp, argv[1],&v.i) != TCL_OK) {
		    return TCL_ERROR;
		}
		if(v.i) flag |= PvmTaskTrace; else flag &= ~PvmTaskTrace;
	    } else {
		Tcl_AppendResult(interp, "invalid option \"",*argv,"\"",(char *)NULL);
		pvmerror = -1; goto pvmend;
	    }
	    argc-=2; argv+=2;
	}
	if(!argc) {
		Tcl_AppendResult(interp,"task to be spawned missing",(char *)NULL);
		pvmerror = -1; goto pvmend;
	}
	if(ntask>0) {
	    int oldtid, oldmsgtag;
	    int n; PvmTidInfo *infoPtr;
	    tids = (int *)ckalloc(ntask*sizeof(int));
	    if(output) {
		oldtid = pvm_setopt(PvmOutputTid,mytid);
		oldmsgtag = pvm_setopt(PvmOutputCode,REDIRECT_MSGTAG);
	    }
	    pvmerror = ntask = pvm_spawn(argv[0],argv+1,flag, where, ntask, tids);
	    if(output) {
		pvm_setopt(PvmOutputTid,oldtid);
		pvm_setopt(PvmOutputCode,oldmsgtag);
	    }
	    if(pvmerror==0) {
		Tcl_AppendResult(interp,"Cannot spawn task \"",argv[0],"\"",(char *)NULL);
		pvmerror = -1; goto pvmend;
	    }
	    if(pvmerror<0) goto pvmend;
	    pvm_notify(PvmTaskExit,KILLTID_MSGTAG, ntask, tids);
	    for(t=tids;ntask>0;t++) {
		if(t!=tids) Tcl_AppendResult(interp," ",(char *)NULL);
		sprintf(storage,"0x%x",*t);
		Tcl_AppendResult(interp,storage,(char *)NULL);
		infoPtr = (PvmTidInfo *) ckalloc(sizeof(PvmTidInfo));
		infoPtr->tid     = *t;
		infoPtr->interp  = interp;
		infoPtr->output  = output;
		infoPtr->nextPtr = firstPvmTidPtr;
		firstPvmTidPtr = infoPtr;
		ntask--;
	    }
	    ckfree(tids);
	}
	pvmerror = 0;
    } else if(!strcmp(argv[1],"kill")) {
	int b;
	if(argc<3) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
				" kill process\"",(char *) NULL);
	    goto pvmend;
	}
	argc-=2; argv+=2;
	pvmerror = 0;
	while(argc>0) {
	    if (!strcmp(*argv,"parent")) {
		tid = pvm_parent();
		if(tid==PvmSysErr||tid==PvmNoParent) {
		    pvmerror = tid;
		    goto pvmend;
		}
	    } else if (!strcmp(*argv,"id")) {
		tid = pvm_mytid();
		if(tid==PvmSysErr) {
		    pvmerror = tid;
		    goto pvmend;
		}
	    } else if(Tcl_GetInt(interp, *argv,&tid) != TCL_OK) {
		goto pvmend;
	    }
	    printf("killing %s\n",*argv);
	    if((v.i = pvm_kill(tid))<0) pvmerror = v.i;
	    argc--; argv++;
	}
    } else if(!strcmp(argv[1],"send")) {
	char *groupname = NULL;
	if(argc<5) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
				" send tid msgtag ....\"",(char *) NULL);
	    goto pvmend;
	}
	if (!strcmp(argv[2],"parent")) {
	    tid = pvm_parent();
	    if(tid==PvmSysErr||tid==PvmNoParent) {
		pvmerror = tid;
		goto pvmend;
	    }
	} else if (!strcmp(argv[2],"id")) {
	    tid = pvm_mytid();
	    if(tid==PvmSysErr) {
		pvmerror = tid;
		goto pvmend;
	    }
	} else {
	    char *p;
	    errno = 0;
	    tid = strtol(argv[2],&p,0);
	    if(p[0] || errno) {
		groupname = argv[2];
	    }
	}
	if (Tcl_GetInt(interp, argv[3], &msgtag) != TCL_OK) {
	    goto pvmend;
	}
	argc-=4; argv+=4;
	pvm_initsend(PvmDataDefault);
	while(argc) {
	    char *p;
	    errno = 0;
	    v.i = strtol(argv[0], &p, 0); /* try integer */
	    if (!p[0] && !errno) {
		pvmerror = pvm_pkint(&v.i,1,1);
	    } else {
		v.d = strtod(argv[0], &p); /* try double */
		if (!p[0] && !errno) {
		    pvmerror = pvm_pkdouble(&v.d,1,1);
		} else {
		    pvmerror = pvm_pkstr(argv[0]);
		}
	    }
	    argc--; argv++;
	    if(pvmerror<0) goto pvmend;
	}
	if(groupname) {
	    pvmerror = pvm_bcast(groupname,msgtag);
	} else {
	    pvmerror = pvm_send(tid, msgtag);
	}
    } else if(!strcmp(argv[1],"recv")) {
	if(argc<3) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
				" recv ....\"",(char *) NULL);
	    goto pvmend;
	}
	argc-=2; argv+=2;
	while(argc) {
	    if(*(interp->result)!='\0')
		Tcl_AppendResult(interp," ",(char *)NULL);
	    if(!strcmp(*argv,"byte")) {
		pvmerror = pvm_upkbyte(&v.c,1,1);
		sprintf(storage,"%d",v.c);
	    } else if(!strcmp(*argv,"char")) {
		pvmerror = pvm_upkbyte(&v.c,1,1);
		sprintf(storage,"%c",v.c);
	    } else if(!strcmp(*argv,"int")) {
		pvmerror = pvm_upkint(&v.i,1,1);
		sprintf(storage,"%d",v.i);
	    } else if(!strcmp(*argv,"tid")) {
		pvmerror = pvm_upkint(&v.i,1,1);
		sprintf(storage,"0x%x",v.i);
	    } else if(!strcmp(*argv,"float")) {
		pvmerror = pvm_upkfloat(&v.f,1,1);
		Tcl_PrintDouble(interp,v.f,storage);
	    } else if(!strcmp(*argv,"double")) {
		pvmerror = pvm_upkdouble(&v.d,1,1);
		Tcl_PrintDouble(interp,v.d,storage);
	    } else if(!strcmp(*argv,"string")) {
		pvmerror = pvm_upkstr(storage);
	    } else {
		Tcl_AppendResult(interp,"wrong format \"",*argv,
				"\", should be byte, char, int, tid, flout, double or string",
				(char *)NULL);
		pvmerror = -1;
	    }
	    if(pvmerror>=0) {
		Tcl_AppendResult(interp,storage,(char *) NULL);
	    } else {
		goto pvmend;
	    }
	    argc--; argv++;
	}
    } else if(!strcmp(argv[1],"mytid")) {
	tid = pvm_mytid();
	if(tid==PvmSysErr) {
	    pvmerror = PvmSysErr;
	    goto pvmend;
	}
	sprintf(storage,"0x%x",tid);
	Tcl_AppendResult(interp,storage,(char *)NULL);
	pvmerror = 0;
    } else if(!strcmp(argv[1],"parent")) {
	if(argc<3) {
	    tid = pvm_parent();
	    if(tid==PvmSysErr) {
		pvmerror = tid;
		goto pvmend;
	    }
	    if(tid==PvmNoParent) {
		pvmerror = 0;
		goto pvmend;
	    }
	    sprintf(storage,"0x%x",tid);
	    Tcl_AppendResult(interp,storage,(char *)NULL);
	    pvmerror = 1;
	    goto pvmend;
	} if(argc<4) {
	    struct pvmtaskinfo *taskp;
	    int i,tid,ntask;
	    if((pvmerror = pvm_tasks(0, &ntask, &taskp))<0) goto pvmend;
	    if(Tcl_GetInt(interp, argv[2],&tid) != TCL_OK) {
		goto pvmend;
	    }
	    i=0; while(i<ntask && taskp[i].ti_tid!=tid) i++;
	    if(i<ntask && taskp[i].ti_ptid) {
		sprintf(storage,"0x%x",taskp[i].ti_ptid);
		Tcl_AppendResult(interp, storage,(char*)NULL);
	    }
	    pvmerror=0;
	    goto pvmend;	    
	} else {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
				" parent tid\"",(char *) NULL);
	    goto pvmend;
	}
	pvmerror = 0;
    } else if(!strcmp(argv[1],"tasks")) {
	struct pvmtaskinfo *taskp;
	int i,ntask;
	if((pvmerror = pvm_tasks(0, &ntask, &taskp))<0) goto pvmend;
	for(i=0; i<ntask; i++) {
	    sprintf(storage,"0x%x",taskp[i].ti_tid);
	    Tcl_AppendElement(interp,storage);
	}
	pvmerror = 0;
    } else if(!strcmp(argv[1],"exit")) {
	pvm_exit(); mytid=0;
	pvmerror = 0;
    } else if(!strcmp(argv[1],"halt")) {
	pvm_halt();
	exit;
    } else if(!strcmp(argv[1],"addhost")) {
	argv+=2; argc-=2;
	while (argc--){
	    if (pvm_addhosts(argv,1,&pvmerror)<1 || pvmerror<0 ) {
		if(pvmerror>=0) pvmerror = PvmBadParam;
		goto pvmend;
	    }
	    argv++;
	}
	pvmerror = 0;
    } else if(!strcmp(argv[1],"delhost")) {
	pvm_delhosts(argv+2,argc-2,NULL);
	pvmerror = 0;
    } else if(!strcmp(argv[1],"joingroup")) {
	if(argc!=3) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
				" joingroup groupname\"",(char *) NULL);
	    goto pvmend;
	}
	if((pvmerror = pvm_joingroup(argv[2]))>=0) {
	    sprintf(storage,"%d",pvmerror);
	    Tcl_AppendResult(interp,storage,(char *)NULL);
	}
    } else if(!strcmp(argv[1],"leavegroup")) {
	if(argc!=3) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
				" joingroup groupname\"",(char *) NULL);
	    goto pvmend;
	}
	pvmerror = pvm_lvgroup(argv[2]);
    } else {
	Tcl_AppendResult(interp, "Wong option for \"",argv[0],
		"\"-command: should be addhost, bind, delhost, exit, halt, joingroup, kill, leavegroup,",
		" mytid, parent, recv, reset, send, spawn or tasks",(char *) NULL);
    }
    pvmend:
    return Pvm_ReturnError(interp,pvmerror);
}

/*
 *--------------------------------------------------------------
 *
 * Pvm_CreatePvmHandler --
 *
 *	Arrange for a given procedure to be invoked whenever
 *	a given Pvm packet arrives.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	From now on, whenever a PVM packet described by tid
 *	and msgtag arrives, proc will be invoked.
 *	If tid and msgtag are already registered then the old
 *      mask and proc and clientData values will be replaced
 *      with new ones
 *
 *--------------------------------------------------------------
 */

void
Pvm_CreatePvmHandler(tid, msgtag, proc, clientData)
    int tid;			/* task descriptor */
    int msgtag;			/* message task */
    Pvm_PvmProc *proc;		/* Procedure to call for each
				 * selected event. */
    ClientData clientData;	/* Arbitrary data to pass to proc. */
{
    register PvmHandler *pvmPtr,*prevPtr,*Ptr;

    /*
     * Make sure tid and msgtag aren't already registered.  Create a
     * new record in the normal case where there's no existing
     * record.
     */

    for (prevPtr = NULL, pvmPtr = firstPvmHandlerPtr; pvmPtr != NULL;
	    prevPtr = pvmPtr,pvmPtr = pvmPtr->nextPtr) {
	if ((unsigned)pvmPtr->tid > (unsigned)tid || (pvmPtr->tid == tid &&
		(unsigned)pvmPtr->msgtag >= (unsigned)msgtag)) {
	    break;
	}
    }
    if (pvmPtr == NULL||pvmPtr->tid!=tid||pvmPtr->msgtag!=msgtag) {
	Ptr = (PvmHandler *) ckalloc(sizeof(PvmHandler));
	Ptr->tid = tid;
	Ptr->msgtag = msgtag;
	Ptr->nextPtr = pvmPtr;
	Ptr->clientData = (ClientData)NULL;
	if(prevPtr ==  NULL) {
	    firstPvmHandlerPtr = Ptr;
	} else {
	    prevPtr->nextPtr = Ptr;
	}
	pvmPtr = Ptr;
    }
    /*
     * The remainder of the initialization below is done
     * regardless of whether or not this is a new record
     * or a modification of an old one.
     */

    pvmPtr->proc = proc;
    pvmPtr->clientData = clientData;

}

/*
 *--------------------------------------------------------------
 *
 * Pvm_DeletePvmHandler --
 *
 *	Cancel a previously-arranged callback arrangement for
 *	a pvm packet.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	If a callback was previously registered, remove it.
 *
 *--------------------------------------------------------------
 */

void
Pvm_DeletePvmHandler(tid, msgtag)
    int tid;
    int msgtag;
{
    register PvmHandler *pvmPtr,*prevPtr;

    /*
     * Find the entry for the given packet (and return if there
     * isn't one).
     */

    for (prevPtr = NULL, pvmPtr = firstPvmHandlerPtr; ;
	    prevPtr = pvmPtr, pvmPtr = pvmPtr->nextPtr) {
	if (pvmPtr == NULL) {
	    return;
	}
	if (pvmPtr->tid == tid && pvmPtr->msgtag == msgtag) {
	    break;
	}
    }

    /*
     * Clean up information in the callback record.
     */

    if (prevPtr == NULL) {
	firstPvmHandlerPtr = pvmPtr->nextPtr;
    } else {
	prevPtr->nextPtr = pvmPtr->nextPtr;
    }
    ckfree((char *) pvmPtr);

}

/*
 *--------------------------------------------------------------
 *
 * Pvm_File_Handler --
 *
 *	This is executed every time a PVM-packet arrives.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The registered procedure will be executed.
 *
 *--------------------------------------------------------------
 */


int Pvm_File_handle(clientData, mask)
ClientData clientData;
int mask;
{
    int info,tid,msgtag, number;
    register PvmHandler *pvmPtr, *prev_pvmPtr;
    register PvmBindInfo *pevPtr, *prev_pevPtr;
    register PvmTidInfo *ptdPtr, *prev_ptdPtr;
    char s[100];

    if((info = pvm_nrecv(-1,-1))>0) {
	pvm_bufinfo(info,&number,&msgtag,&tid);
	if(tid==0x80000000) {
	    switch(msgtag) {
	    case KILLTID_MSGTAG:
		info = pvm_upkint(&tid,1,1);
		if(info<0) {
		    Tcl_AppendResult(Pvm_interp,"Error receiving data from PVM daemon",(char *)NULL);
		    sprintf(s,"\n     msgtag:%d, error:%d",msgtag,info);
		    Tcl_AddErrorInfo(Pvm_interp,s);
		    Tk_BackgroundError(Pvm_interp);
		    return 0;
		} else {
		    for (pvmPtr = firstPvmHandlerPtr; ;pvmPtr = pvmPtr->nextPtr) {
			if (pvmPtr == NULL) {
			    break;
			}
			if (pvmPtr->tid == tid && pvmPtr->msgtag == -2) {
			    /* execute the 'kill' binding */
			    pvmPtr->proc(tid,-2,0,pvmPtr->clientData);
			    break;
			}
		    }
		    prev_ptdPtr = NULL; ptdPtr = firstPvmTidPtr;
		    while(ptdPtr && ptdPtr->tid !=tid) {
			prev_ptdPtr = ptdPtr;
			ptdPtr = ptdPtr->nextPtr;
		    }
		    if(ptdPtr) {
			if(prev_ptdPtr) {
			    prev_ptdPtr->nextPtr = ptdPtr->nextPtr;
			} else {
			    firstPvmTidPtr = ptdPtr->nextPtr;
			}
		    }
		    prev_pvmPtr=NULL; pvmPtr = firstPvmHandlerPtr;
		    while (pvmPtr) {
			if(pvmPtr->tid==tid) {
			    if(prev_pvmPtr) {
				prev_pvmPtr->nextPtr = pvmPtr->nextPtr;
			    } else {
				firstPvmHandlerPtr = pvmPtr->nextPtr;
			    }
			} else {
			    prev_pvmPtr = pvmPtr;
			}
			pvmPtr = pvmPtr->nextPtr;
		    }
		    prev_pevPtr=NULL; pevPtr = firstPvmBindPtr;
		    while (pevPtr) {
			if(pevPtr->tid==tid) {
			    if(prev_pevPtr) {
				prev_pevPtr->nextPtr = pevPtr->nextPtr;
			    } else {
				firstPvmBindPtr = pevPtr->nextPtr;
			    }
			} else {
			    prev_pevPtr = pevPtr;
			}
			pevPtr = pevPtr->nextPtr;
		    }
		    if (!pvm_idle_flag) {
			Tk_DoWhenIdle(Pvm_Cleanup,(ClientData) NULL);
			pvm_idle_flag = 1;
		    }
		}
		break;
	    case REDIRECT_MSGTAG:
		{
		    char *p;
		    int t,c;
		    PvmTidInfo *infoPtr;

		    pvm_upkint(&t,1,1); pvm_upkint(&c,1,1);
		    p = ckalloc(c+1);
		    pvm_upkbyte(p,c,1); p[c] = 0;
/*		    printf("Redirect task: 0x%x string \"%s\"\n",t,p);*/
		    infoPtr = firstPvmTidPtr;
		    while(infoPtr && infoPtr->tid !=t) {
			infoPtr = infoPtr->nextPtr;
		    }
		    if(c && infoPtr && infoPtr->output) {
			fputs(p,infoPtr->output); /* change this to fwrite later */
		    } else if(c) {
/*			printf("EEERROR\n");*/
		    }
		    ckfree(p);
		}
		break;
	    default:
		Tcl_AppendResult(Pvm_interp,"Unknown data received from PVM daemon",(char *)NULL);
		sprintf(s,"\n     msgtag:%d, number: %d",msgtag,number);
		Tcl_AddErrorInfo(Pvm_interp,s);
		Tk_BackgroundError(Pvm_interp);
		return 0;
	    }
	} else {
	    for (pvmPtr = firstPvmHandlerPtr; ;pvmPtr = pvmPtr->nextPtr) {
		if (pvmPtr == NULL) {
		    Tcl_AppendResult(Pvm_interp,"Unknown PVM-packet received",(char *)NULL);
		    sprintf(s,"\n\ttid:\t0x%x\n\tmsgtag:\t%d\n\tnumber:\t%d",tid,msgtag,number);
		    Tcl_AddErrorInfo(Pvm_interp,s);
		    Tk_BackgroundError(Pvm_interp);
		    return 0;
		}
		if ((pvmPtr->tid == -1 ||pvmPtr->tid == tid) &&
		    (pvmPtr->msgtag == -1 ||pvmPtr->msgtag == msgtag)) {
		    break;
		}
	    }
	    /* execute the procedure */
	    pvmPtr->proc(tid,msgtag,number,pvmPtr->clientData);
	}
    }
    return 1;
}

/*
 *--------------------------------------------------------------
 *
 * Pvm_EvalCommand --
 *
 *	This is executed every time a PVM-packet arrives.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The registered procedure will be executed.
 *
 *--------------------------------------------------------------
 */


void Pvm_EvalCommand(tid, msgtag, number, clientData)
int tid;
int msgtag;
int number;
ClientData clientData;
{
    register char *s;
    PvmBindInfo *pevPtr = (PvmBindInfo *) clientData;
    char storage[15];
    char *before = pevPtr->command;
    Tcl_DString ds;
    /*
     *    expand %t and %m
     */
    Tcl_DStringInit(&ds);
    while(1) {
	/* search for first '%'-sign */
	for (s = before ; (*s!=0)&&(*s!='%'); s++) ;
	if(s!=before) {
	    Tcl_DStringAppend(&ds, before, s-before);
	    before = s;
	}
	if (*before == 0) {
	    break;
	}
	/* There's a percent sequence here. Process it. */
	switch (before[1]) {
	    case 't':
		sprintf(storage,"0x%x",tid);
		break;
	    case 'm':
		if(msgtag==-2) {
		    sprintf(storage,"kill");
		} else {
		    sprintf(storage,"%d",msgtag);
		}
		break;
	    case 'n':
		sprintf(storage,"%d",number);
		break;
	    default:
		sprintf(storage,"%%%c",before[1]);
		break;
	}
	before+=2;
	Tcl_DStringAppend(&ds,storage,strlen(storage));
    }
    if(Tcl_GlobalEval(pevPtr->interp,Tcl_DStringValue(&ds))!=TCL_OK) {
	Tcl_AddErrorInfo(pevPtr->interp,"\n    (command bound to PVM)");
	Tk_BackgroundError(pevPtr->interp);
    }
    Tcl_DStringFree(&ds);
}

/*
 *--------------------------------------------------------------
 *
 * PvmEventsPending --
 *
 *	check if there are any more pending Pvm events
 *
 * Results:
 *	1 if there are, 0 if not
 *
 * Side effects:
 *	none
 *
 *--------------------------------------------------------------
 */
int PvmEventsPending()
{
    return (firstPvmBindPtr!=0);
}
